@@ -1,6 +1,8 @@
-# -*- coding: utf-8 -*-
+# -*- coding: future_annotations -*-
+import ast
 import inspect
 import itertools
+import logging
 from typing import TYPE_CHECKING
 import weakref
 
@@ -12,25 +14,26 @@ except ImportError:
 
 from nbsafety.analysis import AttrSubSymbolChain, CallPoint
 from nbsafety.data_model.data_symbol import DataSymbol, DataSymbolType
+from nbsafety.singletons import nbs, nbs_check_init
 
 if TYPE_CHECKING:
     from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, Union
-    import ast
-    from nbsafety.safety import NotebookSafety
     from nbsafety.types import SupportedIndexType
 
 
-class Scope(object):
+logger = logging.getLogger(__name__)
+logger.setLevel(logging.WARNING)
+
+
+class Scope:
     GLOBAL_SCOPE_NAME = '<module>'
 
     def __init__(
-            self,
-            safety: 'NotebookSafety',
-            scope_name: str = GLOBAL_SCOPE_NAME,
-            parent_scope: 'Optional[Scope]' = None,
+        self,
+        scope_name: str = GLOBAL_SCOPE_NAME,
+        parent_scope: Optional[Scope] = None,
     ):
-        self.safety = safety
-        self.scope_name = scope_name
+        self.scope_name = str(scope_name)
         self.parent_scope = parent_scope  # None iff this is the global scope
         self._data_symbol_by_name: Dict[SupportedIndexType, DataSymbol] = {}
 
@@ -59,30 +62,30 @@ class Scope(object):
             return self.parent_scope.non_namespace_parent_scope
         return self.parent_scope
 
-    def make_child_scope(self, scope_name, obj_id=None) -> 'Scope':
+    def make_child_scope(self, scope_name, obj_id=None) -> Scope:
         if obj_id is None:
-            return Scope(self.safety, scope_name, parent_scope=self)
+            return Scope(scope_name, parent_scope=self)
         else:
-            return NamespaceScope(obj_id, self.safety, scope_name, parent_scope=self)
+            return NamespaceScope(obj_id, scope_name, parent_scope=self)
 
-    def put(self, name: 'SupportedIndexType', val: DataSymbol):
+    def put(self, name: SupportedIndexType, val: DataSymbol):
         self._data_symbol_by_name[name] = val
         val.containing_scope = self
 
-    def lookup_data_symbol_by_name_this_indentation(self, name) -> 'Optional[DataSymbol]':
+    def lookup_data_symbol_by_name_this_indentation(self, name, **_) -> Optional[DataSymbol]:
         return self._data_symbol_by_name.get(name, None)
 
     def all_data_symbols_this_indentation(self):
         return self._data_symbol_by_name.values()
 
-    def lookup_data_symbol_by_name(self, name) -> 'Optional[DataSymbol]':
-        ret = self.lookup_data_symbol_by_name_this_indentation(name)
+    def lookup_data_symbol_by_name(self, name, **kwargs) -> Optional[DataSymbol]:
+        ret = self.lookup_data_symbol_by_name_this_indentation(name, **kwargs)
         if ret is None and self.non_namespace_parent_scope is not None:
-            ret = self.non_namespace_parent_scope.lookup_data_symbol_by_name(name)
+            ret = self.non_namespace_parent_scope.lookup_data_symbol_by_name(name, **kwargs)
         return ret
 
     @staticmethod
-    def _get_name_to_obj_mapping(obj, dc) -> 'Dict[SupportedIndexType, Any]':
+    def _get_name_to_obj_mapping(obj, dc) -> Dict[SupportedIndexType, Any]:
         if obj is None:
             return get_ipython().ns_table['user_global']
         elif dc is not None and dc.is_subscript:
@@ -126,75 +129,102 @@ class Scope(object):
                 obj = Scope._get_name_to_obj_mapping(obj, dsym)[name]
             except (KeyError, IndexError, Exception):
                 break
-            cur_scope = self.safety.namespaces.get(id(obj), None)
+            cur_scope = nbs().namespaces.get(id(obj), None)
             if cur_scope is None:
                 break
         else:
             success = True
         return dsym, next_dsym, success
 
-    def upsert_data_symbol_for_name(
-            self,
-            name: 'Union[str, int]',
-            obj: 'Any',
-            deps: 'Set[DataSymbol]',
-            stmt_node: 'ast.AST',
-            is_subscript,
-            overwrite=True,
-            is_function_def=False,
-            is_import=False,
-            class_scope: 'Optional[Scope]' = None,
-            propagate=True
-    ) -> 'DataSymbol':
-        dc, old_dc, old_id = self._upsert_data_symbol_for_name_inner(
-            name, obj, deps, stmt_node, is_subscript,
-            overwrite=overwrite, is_function_def=is_function_def, is_import=is_import, class_scope=class_scope
-        )
-        # print(self, 'upsert', name, 'with deps', deps, 'and children', list(dc.children_by_cell_position.values()))
-        dc.update_deps(deps, overwrite=overwrite, propagate=propagate)
-        return dc
-
-    def _upsert_data_symbol_for_name_inner(
-            self,
-            name: 'Union[str, int]',
-            obj: 'Any',
-            deps: 'Set[DataSymbol]',
-            stmt_node: 'ast.AST',
-            is_subscript,
-            overwrite=True,
-            is_function_def=False,
-            is_import=False,
-            class_scope: 'Optional[Scope]' = None,
-    ) -> 'Tuple[DataSymbol, Optional[DataSymbol], Optional[int]]':
-        # print(self, 'upsert', name)
+    @staticmethod
+    def _resolve_symbol_type(
+        overwrite: bool = True,
+        is_subscript: bool = False,
+        is_function_def: bool = False,
+        is_import: bool = False,
+        is_anonymous: bool = False,
+        class_scope: Optional[Scope] = None,
+    ):
         assert not (class_scope is not None and (is_function_def or is_import))
-        symbol_type = DataSymbolType.DEFAULT
         if is_function_def:
             assert overwrite
             assert not is_subscript
-            symbol_type = DataSymbolType.FUNCTION
-            # return self._upsert_function_data_symbol_for_name(name, obj, deps)
+            return DataSymbolType.FUNCTION
         elif is_import:
             assert overwrite
             assert not is_subscript
-            symbol_type = DataSymbolType.IMPORT
+            return DataSymbolType.IMPORT
         elif class_scope is not None:
             assert overwrite
             assert not is_subscript
-            symbol_type = DataSymbolType.CLASS
-            # return self._upsert_class_data_symbol_for_name(name, obj, deps, class_scope)
+            return DataSymbolType.CLASS
         elif is_subscript:
-            symbol_type = DataSymbolType.SUBSCRIPT
+            return DataSymbolType.SUBSCRIPT
+        elif is_anonymous:
+            return DataSymbolType.ANONYMOUS
+        else:
+            return DataSymbolType.DEFAULT
+
+    def upsert_data_symbol_for_name(
+        self,
+        name: SupportedIndexType,
+        obj: Any,
+        deps: Set[DataSymbol],
+        stmt_node: ast.AST,
+        overwrite: bool = True,
+        is_subscript: bool = False,
+        is_function_def: bool = False,
+        is_import: bool = False,
+        is_anonymous: bool = False,
+        class_scope: Optional[Scope] = None,
+        symbol_type: Optional[DataSymbolType] = None,
+        propagate: bool = True,
+        implicit: bool = False,
+    ) -> DataSymbol:
+        symbol_type = symbol_type or self._resolve_symbol_type(
+            overwrite=overwrite,
+            is_subscript=is_subscript,
+            is_function_def=is_function_def,
+            is_import=is_import,
+            is_anonymous=is_anonymous,
+            class_scope=class_scope
+        )
+        dsym, old_dsym, old_id = self._upsert_data_symbol_for_name_inner(
+            name,
+            obj,
+            deps,
+            symbol_type,
+            stmt_node,
+            implicit=implicit,
+        )
+        dsym.update_deps(deps, overwrite=overwrite, propagate=propagate)
+        return dsym
+
+    def _upsert_data_symbol_for_name_inner(
+        self,
+        name: SupportedIndexType,
+        obj: Any,
+        deps: Set[DataSymbol],
+        symbol_type: DataSymbolType,
+        stmt_node: ast.AST,
+        implicit: bool = False,
+    ) -> Tuple[DataSymbol, Optional[DataSymbol], Optional[int]]:
         old_id = None
-        old_dc = self.lookup_data_symbol_by_name_this_indentation(name)
-        if old_dc is not None and self.is_globally_accessible:
-            old_id = old_dc.cached_obj_id
+        old_dsym = self.lookup_data_symbol_by_name_this_indentation(
+            name, is_subscript=symbol_type == DataSymbolType.SUBSCRIPT, skip_cloned_lookup=True,
+        )
+        if implicit and symbol_type != DataSymbolType.ANONYMOUS:
+            assert old_dsym is None, 'expected None, got %s' % old_dsym
+        if old_dsym is not None and self.is_globally_accessible:
+            old_id = old_dsym.cached_obj_id
             # TODO: handle case where new dc is of different type
-            if name in self.data_symbol_by_name(old_dc.is_subscript) and old_dc.symbol_type == symbol_type:
-                old_dc.update_obj_ref(obj, refresh_cached=False)
-                # old_dc.update_type(symbol_type)
-                old_dc.update_stmt_node(stmt_node)
-                return old_dc, old_dc, old_id
+            if name in self.data_symbol_by_name(old_dsym.is_subscript) and old_dsym.symbol_type == symbol_type:
+                old_dsym.update_obj_ref(obj, refresh_cached=False)
+                # old_dsym.update_type(symbol_type)
+                # if we're updating a pre-existing one, it should not be an implicit upsert
+                assert stmt_node is not None
+                old_dsym.update_stmt_node(stmt_node)
+                return old_dsym, old_dsym, old_id
             else:
                 # In this case, we are copying from a class and we need the dsym from which we are copying
                 # as able to propagate to the new dsym.
@@ -202,9 +232,9 @@ class Scope(object):
                 # class Foo:
                 #     shared = 99
                 # foo = Foo()
-                # foo.shared = 42  # old_dc refers to Foo.shared here
+                # foo.shared = 42  # old_dsym refers to Foo.shared here
                 # Earlier, we were explicitly adding Foo.shared as a dependency of foo.shared as follows:
-                # deps.add(old_dc)
+                # deps.add(old_dsym)
                 # But it turns out not to be necessary because foo depends on Foo, and changing Foo.shared will
                 # propagate up the namespace hierarchy to Foo, which propagates to foo, which then propagates to
                 # all of foo's namespace children (e.g. foo.shared).
@@ -214,10 +244,22 @@ class Scope(object):
                 # If we do this, then we should go back to explicitly adding the dep as follows:
                 # EDIT: added check to avoid propagating along class -> instance edge when class not redefined, so now
                 # it is important to explicitly add this dep.
-                deps.add(old_dc)
-        dc = DataSymbol(name, symbol_type, obj, self, self.safety, stmt_node=stmt_node, parents=deps, refresh_cached_obj=False)
-        self.put(name, dc)
-        return dc, old_dc, old_id
+                deps.add(old_dsym)
+        if isinstance(self, NamespaceScope) and symbol_type == DataSymbolType.DEFAULT and self.cloned_from is not None:
+            # add the cloned symbol as a dependency of the symbol about to b ecreated
+            new_dep = self.cloned_from.lookup_data_symbol_by_name_this_indentation(name, is_subscript=False)
+            if new_dep is not None:
+                deps.add(new_dep)
+        dsym = DataSymbol(
+            name, symbol_type, obj, self, stmt_node=stmt_node, parents=deps, refresh_cached_obj=False, implicit=implicit
+        )
+        self.put(name, dsym)
+        return dsym, old_dsym, old_id
+
+    def delete_data_symbol_for_name(self, name: SupportedIndexType, is_subscript: bool = False):
+        dsym = self._data_symbol_by_name.pop(name, None)
+        if dsym is not None:
+            dsym.update_deps(set())
 
     @property
     def is_global(self):
@@ -238,7 +280,7 @@ class Scope(object):
         return self.parent_scope.global_scope
 
     @property
-    def full_path(self) -> 'Tuple[str, ...]':
+    def full_path(self) -> Tuple[str, ...]:
         path = (self.scope_name,)
         if self.is_global:
             return path
@@ -261,14 +303,16 @@ class Scope(object):
         else:
             return self.scope_name
 
-    def make_namespace_qualified_name(self, dc: 'DataSymbol') -> str:
+    def make_namespace_qualified_name(self, dc: DataSymbol) -> str:
         return str(dc.name)
 
 
 class NamespaceScope(Scope):
+    ANONYMOUS = '<anonymous_namespace>'
+
     # TODO: support (multiple) inheritance by allowing
     #  NamespaceScopes from classes to clone their parent class's NamespaceScopes
-    def __init__(self, obj: 'Any', *args, **kwargs):
+    def __init__(self, obj: Any, *args, **kwargs):
         super().__init__(*args, **kwargs)
         self.cloned_from: Optional[NamespaceScope] = None
         self.child_clones: List[NamespaceScope] = []
@@ -276,30 +320,63 @@ class NamespaceScope(Scope):
         self._tombstone = tombstone
         self._obj_ref = obj_ref
         self.obj_id = obj_id
-        self.safety.namespaces[obj_id] = self
+        nbs().namespaces[obj_id] = self
         self.max_defined_timestamp = 0
         self._subscript_data_symbol_by_name: Dict[SupportedIndexType, DataSymbol] = {}
 
+    def __bool__(self):
+        # in order to override if __len__ returns 0
+        return True
+
+    def __len__(self):
+        obj = self._obj_ref()
+        if not isinstance(obj, (dict, list, tuple)):
+            raise TypeError("tried to get length of non-container namespace %s: %s", self, obj)
+        return len(obj)
+
+    def _iter_inner(self, obj):
+        for i in range(len(obj)):
+            yield self.lookup_data_symbol_by_name_this_indentation(i, is_subscript=True)
+
+    def __iter__(self):
+        obj = self._obj_ref()
+        if not isinstance(obj, (list, tuple)):
+            raise TypeError("tried to iterate through non-sequence namespace %s: %s", self, obj)
+        # do the validation before starting the generator part so that we raise immediately
+        return self._iter_inner(obj)
+
+    def _items_inner(self, obj):
+        for key in obj.keys():
+            yield key, self.lookup_data_symbol_by_name_this_indentation(key, is_subscript=True)
+
+    def items(self):
+        obj = self._obj_ref()
+        if not isinstance(obj, dict):
+            raise TypeError("tried to get iterate through items of non-dict namespace: %s", obj)
+        # do the validation before starting the generator part so that we raise immediately
+        return self._items_inner(obj)
+
     @property
     def is_garbage(self):
-        return self._tombstone or self.obj_id not in self.safety.aliases or self.obj_id not in self.safety.namespaces
+        return self._tombstone or self.obj_id not in nbs().aliases or self.obj_id not in nbs().namespaces
 
     @property
     def is_subscript(self):
-        try:
-            return next(iter(self.safety.aliases.get(self.obj_id, None))).is_subscript
-        except (StopIteration, TypeError):
+        dsym = nbs().get_first_full_symbol(self.obj_id)
+        if dsym is None:
             return False
+        else:
+            return dsym.is_subscript
 
     def update_obj_ref(self, obj):
         tombstone, obj_ref, obj_id = self._update_obj_ref_inner(obj)
         self._tombstone = tombstone
         self._obj_ref = obj_ref
         self.obj_id = obj_id
-        self.safety.namespaces[obj_id] = self
+        nbs().namespaces[obj_id] = self
 
     def clear_namespace(self, prev_obj_id):
-        if prev_obj_id != self.obj_id and prev_obj_id in self.safety.namespaces:
+        if prev_obj_id != self.obj_id and prev_obj_id in nbs().namespaces:
             raise ValueError('precondition failed; namespace should no longer be registered before we can clear')
         self._data_symbol_by_name.clear()
         self._subscript_data_symbol_by_name.clear()
@@ -309,13 +386,17 @@ class NamespaceScope(Scope):
         try:
             obj_ref = weakref.ref(obj, self._obj_reference_expired_callback)
         except TypeError:
-            obj_ref = None
+            obj_ref = lambda: obj
         obj_id = id(obj)
         return tombstone, obj_ref, obj_id
 
     def _obj_reference_expired_callback(self, *_):
         self._tombstone = True
-        self.safety.garbage_namespace_obj_ids.add(self.obj_id)
+        safety = nbs_check_init()
+        if safety is None:
+            # can happen e.g. if program is exiting
+            return
+        safety.garbage_namespace_obj_ids.add(self.obj_id)
 
     def data_symbol_by_name(self, is_subscript=False):
         if is_subscript:
@@ -323,8 +404,8 @@ class NamespaceScope(Scope):
         else:
             return self._data_symbol_by_name
 
-    def clone(self, obj: 'Any'):
-        cloned = NamespaceScope(obj, self.safety)
+    def clone(self, obj: Any):
+        cloned = NamespaceScope(obj, nbs())
         cloned.__dict__ = dict(self.__dict__)
         cloned.cloned_from = self
         cloned.update_obj_ref(obj)
@@ -333,7 +414,10 @@ class NamespaceScope(Scope):
         self.child_clones.append(cloned)
         return cloned
 
-    def make_namespace_qualified_name(self, dc: 'DataSymbol'):
+    def fresh_copy(self, obj: Any):
+        return NamespaceScope(obj, self.scope_name, self.parent_scope)
+
+    def make_namespace_qualified_name(self, dc: DataSymbol):
         path = self.full_namespace_path
         name = str(dc.name)
         if path:
@@ -344,7 +428,7 @@ class NamespaceScope(Scope):
         else:
             return name
 
-    def lookup_data_symbol_by_name_this_indentation(self, name, is_subscript=None):
+    def lookup_data_symbol_by_name_this_indentation(self, name, is_subscript=None, skip_cloned_lookup=False):
         # TODO: specify in arguments whether `name` refers to a subscript
         if is_subscript is None:
             ret = self._data_symbol_by_name.get(name, None)
@@ -354,11 +438,25 @@ class NamespaceScope(Scope):
             ret = self._subscript_data_symbol_by_name.get(name, None)
         else:
             ret = self._data_symbol_by_name.get(name, None)
-        if ret is None and self.cloned_from is not None:
-            ret = self.cloned_from.lookup_data_symbol_by_name_this_indentation(name)
+        if not skip_cloned_lookup and ret is None and self.cloned_from is not None and not is_subscript and isinstance(name, str):
+            if name not in getattr(self._obj_ref(), '__dict__', {}):
+                # only fall back to the class sym if it's not present in the corresponding obj for this scope
+                ret = self.cloned_from.lookup_data_symbol_by_name_this_indentation(name, is_subscript=is_subscript)
         return ret
 
-    def all_data_symbols_this_indentation(self, exclude_class=False, is_subscript=None) -> 'Iterable[DataSymbol]':
+    def delete_data_symbol_for_name(self, name: SupportedIndexType, is_subscript: bool = False):
+        logger.info("delete %s from %s", name, self)
+        if is_subscript:
+            dsym = self._subscript_data_symbol_by_name.pop(name, None)
+            if dsym is None and name == -1 and isinstance(self._obj_ref(), list):
+                name = len(self._obj_ref())  # it will have already been deleted, so don't subtract 1
+                dsym = self._subscript_data_symbol_by_name.pop(name, None)
+            if dsym is not None:
+                dsym.update_deps(set(), deleted=True)
+        else:
+            super().delete_data_symbol_for_name(name)
+
+    def all_data_symbols_this_indentation(self, exclude_class=False, is_subscript=None) -> Iterable[DataSymbol]:
         if is_subscript is None:
             dsym_collections_to_chain: List[Iterable] = [
                 self._data_symbol_by_name.values(), self._subscript_data_symbol_by_name.values()
@@ -383,7 +481,7 @@ class NamespaceScope(Scope):
     def num_symbols(self):
         return self.num_dotted_symbols + self.num_subscript_symbols
 
-    def put(self, name: 'SupportedIndexType', val: DataSymbol):
+    def put(self, name: SupportedIndexType, val: DataSymbol):
         if val.is_subscript:
             self._subscript_data_symbol_by_name[name] = val
         else:
@@ -393,9 +491,9 @@ class NamespaceScope(Scope):
         val.containing_scope = self
 
     def refresh(self):
-        self.max_defined_timestamp = self.safety.cell_counter()
+        self.max_defined_timestamp = nbs().cell_counter()
 
-    def get_earliest_ancestor_containing(self, obj_id: int, is_subscript: bool) -> 'Optional[NamespaceScope]':
+    def get_earliest_ancestor_containing(self, obj_id: int, is_subscript: bool) -> Optional[NamespaceScope]:
         # TODO: test this properly
         ret = None
         if self.namespace_parent_scope is not None:
@@ -409,7 +507,7 @@ class NamespaceScope(Scope):
             return None
 
     @property
-    def namespace_parent_scope(self) -> 'Optional[NamespaceScope]':
+    def namespace_parent_scope(self) -> Optional[NamespaceScope]:
         if self.parent_scope is not None and isinstance(self.parent_scope, NamespaceScope):
             return self.parent_scope
         return None
