@@ -1,7 +1,8 @@
-# -*- coding: utf-8 -*-
+# -*- coding: future_annotations -*-
 from ipykernel.ipkernel import IPythonKernel
 from nbsafety.version import __version__
-from nbsafety.safety import NotebookSafety, SafetyRunMode
+from nbsafety.safety import NotebookSafety
+from nbsafety.singletons import nbs
 
 
 class SafeKernel(IPythonKernel):
@@ -10,11 +11,11 @@ class SafeKernel(IPythonKernel):
 
     def __init__(self, **kwargs):
         super().__init__(**kwargs)
-        self._safety = NotebookSafety(use_comm=True)
+        NotebookSafety.instance(use_comm=True)
 
     def do_execute(self, code, silent, store_history=False, user_expressions=None, allow_stdin=False):
         super_ = super()
 
         def _run_cell_func(cell):
             return super_.do_execute(cell, silent, store_history, user_expressions, allow_stdin)
-        return self._safety.safe_execute(code, _run_cell_func)
+        return nbs().safe_execute(code, _run_cell_func)
