@@ -1,8 +1,9 @@
-# -*- coding: utf-8 -*-
+# -*- coding: future_annotations -*-
 from typing import TYPE_CHECKING
+from nbsafety.data_model.data_symbol import DataSymbol
 
 if TYPE_CHECKING:
-    from typing import Any, List
+    from typing import Iterable, List
     from nbsafety.safety import NotebookSafety
 
 
@@ -14,7 +15,7 @@ if TYPE_CHECKING:
 
 USAGE = """Options:
 
-show_dependency <variable_name> <variable_name2> ...: 
+show_[deps|dependencies] <variable_name> <variable_name2> ...: 
     - This will print out the dependencies for given global variables.
       Multiple variables should be separated with spaces.
 
@@ -36,26 +37,30 @@ turn_off_warnings_for  <variable_name> <variable_name2> ...:
       stale dependencies now. Multiple variables should be separated with spaces."""
 
 
-def show_deps(safety: 'NotebookSafety', line: 'List[str]'):
+def show_deps(safety: NotebookSafety, line: List[str]):
     if len(line) == 1:
-        print("Usage: %safety show_dependency <variable_name> <variable_name2> ...")
+        print("Usage: %safety show_[deps|dependencies] <variable_name> <variable_name2> ...")
         return
     for data_sym_name in line[1:]:
         data_sym = safety.global_scope.lookup_data_symbol_by_name(data_sym_name)
+        parents = {par for par in data_sym.parents if not par.is_anonymous}
+        if data_sym.required_cell_num > 0:
+            dsym_extra_info = 'defined {}; required {}'.format(data_sym.defined_cell_num, data_sym.required_cell_num)
+        else:
+            dsym_extra_info = 'defined in cell {}'.format(data_sym.defined_cell_num)
         if data_sym:
-            print("DataSymbol {} (defined {}; required {}) is dependent on {}".format(
+            print("DataSymbol {} ({}) is dependent on {}".format(
                 data_sym.readable_name,
-                data_sym.defined_cell_num,
-                data_sym.required_cell_num,
-                [str(n) for n in data_sym.parents] if data_sym.parents else "Nothing"
+                dsym_extra_info,
+                parents or "nothing"
             ))
         else:
             print("Cannot find DataSymbol", data_sym_name)
 
 
-def show_stale(safety: 'NotebookSafety', line: 'List[str]'):
+def show_stale(safety: NotebookSafety, line: List[str]):
     if len(line) < 2 or line[1] == 'global':
-        dsym_sets: Any = [safety.global_scope.all_data_symbols_this_indentation()]
+        dsym_sets: Iterable[Iterable[DataSymbol]] = [safety.global_scope.all_data_symbols_this_indentation()]
     elif line[1] == 'all':
         dsym_sets = safety.aliases.values()
     else:
@@ -64,24 +69,24 @@ def show_stale(safety: 'NotebookSafety', line: 'List[str]'):
     stale_set = set()
     for dsym_set in dsym_sets:
         for data_sym in dsym_set:
-            if data_sym.is_stale:
+            if data_sym.is_stale and not data_sym.is_anonymous:
                 stale_set.add(data_sym)
     if not stale_set:
-        print("No DataSymbol has stale dependency for now!")
+        print("No DataSymbol has stale dependencies for now!")
     elif len(stale_set) == 1:
-        print("The only DataSymbol with stale dependencies is:", str(stale_set.pop()))
+        print("The only DataSymbol with stale dependencies is:", next(iter(stale_set)))
     else:
-        print("DataSymbols with stale dependencies are:", [str(n) for n in stale_set])
+        print("DataSymbols with stale dependencies are:", stale_set)
 
 
-def trace_messages(safety: 'NotebookSafety', line: 'List[str]'):
+def trace_messages(safety: NotebookSafety, line: List[str]):
     if len(line) != 2:
         print("Usage: %safety trace_messages [enabled|disabled] ...")
         return
-    safety.config.trace_messages_enabled = (line[1].lower().startswith("enable"))
+    safety.settings.trace_messages_enabled = (line[1].lower().startswith("enable"))
 
 
-def remove_dep(safety: 'NotebookSafety', line: 'List[str]'):
+def remove_dep(safety: NotebookSafety, line: List[str]):
     if len(line) != 3:
         print("Usage: %safety remove_dependency <parent_name> <child_name>")
         return
@@ -101,7 +106,7 @@ def remove_dep(safety: 'NotebookSafety', line: 'List[str]'):
     child_data_sym.parents.remove(parent_data_sym)
 
 
-def add_dep(safety: 'NotebookSafety', line: 'List[str]'):
+def add_dep(safety: NotebookSafety, line: List[str]):
     if len(line) != 3:
         print("Usage: %safety add_dependency <parent_name> <child_name>")
         return
@@ -120,7 +125,7 @@ def add_dep(safety: 'NotebookSafety', line: 'List[str]'):
     child_data_sym.parents.add(parent_data_sym)
 
 
-def turn_off_warnings_for(safety: 'NotebookSafety', line: 'List[str]'):
+def turn_off_warnings_for(safety: NotebookSafety, line: List[str]):
     if len(line) <= 1:
         print("Usage: %safety turn_off_warnings_for <variable_name> <variable_name2> ...")
         return
@@ -133,7 +138,7 @@ def turn_off_warnings_for(safety: 'NotebookSafety', line: 'List[str]'):
             print("Cannot find DataSymbol", data_sym_name)
 
 
-def turn_on_warnings_for(safety: 'NotebookSafety', line: 'List[str]'):
+def turn_on_warnings_for(safety: NotebookSafety, line: List[str]):
     if len(line) <= 1:
         print("Usage: %safety turn_on_warnings_for <variable_name> <variable_name2> ...")
         return
