@@ -1,11 +1,11 @@
-# -*- coding: utf-8 -*-
+# -*- coding: future_annotations -*-
 import ast
 from collections import defaultdict
 from contextlib import contextmanager
 import inspect
 import logging
 import re
-from typing import cast, TYPE_CHECKING
+from typing import cast, TYPE_CHECKING, NamedTuple
 
 from IPython import get_ipython
 from IPython.core.magic import register_cell_magic, register_line_magic
@@ -22,15 +22,17 @@ from nbsafety.ipython_utils import (
     save_number_of_currently_executing_cell,
 )
 from nbsafety import line_magics
+from nbsafety.data_model.data_symbol import DataSymbol
 from nbsafety.data_model.scope import Scope, NamespaceScope
 from nbsafety.run_mode import SafetyRunMode
-from nbsafety.tracing import SafetyAstRewriter, TracingManager
-from nbsafety.utils import DotDict
+from nbsafety import singletons
+from nbsafety.tracing.safety_ast_rewriter import SafetyAstRewriter
+from nbsafety.tracing.trace_manager import TraceManager
+# from nbsafety.utils.mixins import EnforceSingletonMixin
 
 if TYPE_CHECKING:
     from typing import Any, Dict, List, Set, Optional, Tuple, Union
     from types import FrameType
-    from nbsafety.data_model.data_symbol import DataSymbol
     CellId = Union[str, int]
 
 logger = logging.getLogger(__name__)
@@ -42,7 +44,7 @@ _SAFETY_LINE_MAGIC = 'safety'
 _NB_MAGIC_PATTERN = re.compile(r'(^%|^!|^cd |\?$)')
 
 
-def _safety_warning(node: 'DataSymbol'):
+def _safety_warning(node: DataSymbol):
     if not node.is_stale:
         raise ValueError('Expected node with stale ancestor; got %s' % node)
     if node.defined_cell_num < 1:
@@ -58,10 +60,26 @@ def _safety_warning(node: 'DataSymbol'):
     )
 
 
-class NotebookSafety(object):
+class NotebookSafetySettings(NamedTuple):
+    store_history: bool
+    test_context: bool
+    use_comm: bool
+    trace_messages_enabled: bool
+    backwards_cell_staleness_propagation: bool
+    track_dependencies: bool
+    naive_refresher_computation: bool
+    skip_unsafe_cells: bool
+    mode: SafetyRunMode
+
+    def get(self, key, default):
+        return getattr(self, key, default)
+
+
+class NotebookSafety(singletons.NotebookSafety):
     """Holds all the state necessary to detect stale dependencies in Jupyter notebooks."""
     def __init__(self, cell_magic_name=None, use_comm=False, **kwargs):
-        self.config = DotDict(dict(
+        super().__init__()
+        self.settings = NotebookSafetySettings(
             store_history=kwargs.pop('store_history', True),
             test_context=kwargs.pop('test_context', False),
             use_comm=use_comm,
@@ -71,19 +89,18 @@ class NotebookSafety(object):
             naive_refresher_computation=False,
             skip_unsafe_cells=kwargs.pop('skip_unsafe', True),
             mode=SafetyRunMode.get(),
-            **kwargs
-        ))
+        )
         # Note: explicitly adding the types helps PyCharm's built-in code inspection
-        self.namespaces: 'Dict[int, NamespaceScope]' = {}
-        self.aliases: 'Dict[int, Set[DataSymbol]]' = defaultdict(set)
-        self.global_scope: 'Scope' = Scope(self)
-        self.updated_symbols: 'Set[DataSymbol]' = set()
-        self.updated_scopes: 'Set[NamespaceScope]' = set()
-        self.garbage_namespace_obj_ids: 'Set[int]' = set()
-        self.ast_node_by_id: 'Dict[int, ast.AST]' = {}
-        self.statement_cache: 'Dict[int, Dict[int, ast.stmt]]' = defaultdict(dict)
-        self.statement_to_func_cell: 'Dict[int, DataSymbol]' = {}
-        self.tracing_manager: 'TracingManager' = TracingManager(self)
+        self.namespaces: Dict[int, NamespaceScope] = {}
+        self.aliases: Dict[int, Set[DataSymbol]] = defaultdict(set)
+        self.global_scope: Scope = Scope()
+        self.updated_symbols: Set[DataSymbol] = set()
+        self.updated_scopes: Set[NamespaceScope] = set()
+        self.garbage_namespace_obj_ids: Set[int] = set()
+        self.ast_node_by_id: Dict[int, ast.AST] = {}
+        self.statement_cache: Dict[int, Dict[int, ast.stmt]] = defaultdict(dict)
+        self.cell_content_by_counter: Dict[int, str] = {}
+        self.statement_to_func_cell: Dict[int, DataSymbol] = {}
         self.stale_dependency_detected = False
         self.active_cell_position_idx = -1
         self._last_execution_counter = 0
@@ -98,35 +115,45 @@ class NotebookSafety(object):
         self._prev_cell_stale_symbols: Set[DataSymbol] = set()
         self._cell_counter = 1
         self._recorded_cell_name_to_cell_num = True
-        self._cell_name_to_cell_num_mapping: 'Dict[str, int]' = {}
-        self._ast_transformer_raised: 'Optional[Exception]' = None
+        self._cell_name_to_cell_num_mapping: Dict[str, int] = {}
+        self._ast_transformer_raised: Optional[Exception] = None
         if use_comm:
             get_ipython().kernel.comm_manager.register_target(__package__, self._comm_target)
 
     @property
     def is_develop(self) -> bool:
-        return self.config.get('mode', SafetyRunMode.PRODUCTION) == SafetyRunMode.DEVELOP
+        return self.settings.get('mode', SafetyRunMode.PRODUCTION) == SafetyRunMode.DEVELOP
 
     @property
     def is_test(self) -> bool:
-        return self.config.get('test_context', False)
+        return self.settings.get('test_context', False)
+
+    def get_first_full_symbol(self, obj_id: int) -> Optional[DataSymbol]:
+        for alias in self.aliases.get(obj_id, []):
+            if not alias.is_anonymous:
+                return alias
+        return None
 
     def cell_counter(self):
-        if self.config.store_history:
+        if self.settings.store_history:
             return cell_counter()
         else:
             return self._cell_counter
 
-    def set_ast_transformer_raised(self, new_val: 'Optional[Exception]' = None) -> 'Optional[Exception]':
+    def set_ast_transformer_raised(self, new_val: Optional[Exception] = None) -> Optional[Exception]:
         ret = self._ast_transformer_raised
         self._ast_transformer_raised = new_val
         return ret
 
-    def get_position(self, frame: 'FrameType'):
-        cell_num = self._cell_name_to_cell_num_mapping[frame.f_code.co_filename.split('-')[3]]
-        return cell_num, frame.f_lineno
+    def get_position(self, frame: FrameType):
+        try:
+            cell_num = self._cell_name_to_cell_num_mapping[frame.f_code.co_filename.split('-')[3]]
+            return cell_num, frame.f_lineno
+        except KeyError as e:
+            print(frame.f_code.co_filename)
+            raise e
 
-    def maybe_set_name_to_cell_num_mapping(self, frame: 'FrameType'):
+    def maybe_set_name_to_cell_num_mapping(self, frame: FrameType):
         if self._recorded_cell_name_to_cell_num:
             return
         self._recorded_cell_name_to_cell_num = True
@@ -152,7 +179,7 @@ class NotebookSafety(object):
             if cell_id is not None:
                 self._counters_by_cell_id[cell_id] = self._last_execution_counter
             cells_by_id = request['content_by_cell_id']
-            if self.config.get('backwards_cell_staleness_propagation', True):
+            if self.settings.get('backwards_cell_staleness_propagation', True):
                 order_index_by_id = None
                 last_cell_exec_position_idx = -1
             else:
@@ -167,14 +194,14 @@ class NotebookSafety(object):
             logger.error('Unsupported request type for request %s' % request)
 
     def check_and_link_multiple_cells(
-            self,
-            cells_by_id: 'Dict[CellId, str]',
-            order_index_by_cell_id: 'Optional[Dict[CellId, int]]' = None
-    ) -> 'Dict[str, Any]':
+        self,
+        cells_by_id: Dict[CellId, str],
+        order_index_by_cell_id: Optional[Dict[CellId, int]] = None
+    ) -> Dict[str, Any]:
         stale_cells = set()
         fresh_cells = []
-        stale_symbols_by_cell_id: 'Dict[CellId, Set[DataSymbol]]' = {}
-        killing_cell_ids_for_symbol: 'Dict[DataSymbol, Set[CellId]]' = defaultdict(set)
+        stale_symbols_by_cell_id: Dict[CellId, Set[DataSymbol]] = {}
+        killing_cell_ids_for_symbol: Dict[DataSymbol, Set[CellId]] = defaultdict(set)
         for cell_id, cell_content in cells_by_id.items():
             if (order_index_by_cell_id is not None and
                     order_index_by_cell_id.get(cell_id, -1) <= self.active_cell_position_idx):
@@ -192,11 +219,11 @@ class NotebookSafety(object):
                     killing_cell_ids_for_symbol[dead_sym].add(cell_id)
             except SyntaxError:
                 continue
-        stale_links: 'Dict[CellId, Set[CellId]]' = defaultdict(set)
-        refresher_links: 'Dict[CellId, List[CellId]]' = defaultdict(list)
+        stale_links: Dict[CellId, Set[CellId]] = defaultdict(set)
+        refresher_links: Dict[CellId, List[CellId]] = defaultdict(list)
         for stale_cell_id in stale_cells:
             stale_syms = stale_symbols_by_cell_id[stale_cell_id]
-            if self.config.get('naive_refresher_computation', False):
+            if self.settings.get('naive_refresher_computation', False):
                 refresher_cell_ids = self._naive_compute_refresher_cells(
                     stale_cell_id,
                     stale_syms,
@@ -235,12 +262,12 @@ class NotebookSafety(object):
         }
 
     def _naive_compute_refresher_cells(
-            self,
-            stale_cell_id: 'CellId',
-            stale_symbols: 'Set[DataSymbol]',
-            cells_by_id: 'Dict[CellId, str]',
-            order_index_by_cell_id: 'Optional[Dict[CellId, int]]' = None
-    ) -> 'Set[CellId]':
+        self,
+        stale_cell_id: CellId,
+        stale_symbols: Set[DataSymbol],
+        cells_by_id: Dict[CellId, str],
+        order_index_by_cell_id: Optional[Dict[CellId, int]] = None
+    ) -> Set[CellId]:
         refresher_cell_ids: Set[CellId] = set()
         stale_cell_content = cells_by_id[stale_cell_id]
         for cell_id, cell_content in cells_by_id.items():
@@ -269,7 +296,7 @@ class NotebookSafety(object):
                 lines.append(line)
         return ast.parse('\n'.join(lines))
 
-    def _get_max_defined_cell_num_for_symbols(self, symbols: 'Set[DataSymbol]') -> int:
+    def _get_max_defined_cell_num_for_symbols(self, symbols: Set[DataSymbol]) -> int:
         max_defined_cell_num = -1
         for dsym in symbols:
             max_defined_cell_num = max(max_defined_cell_num, dsym.defined_cell_num)
@@ -279,9 +306,9 @@ class NotebookSafety(object):
         return max_defined_cell_num
 
     def _check_cell_and_resolve_symbols(
-            self,
-            cell: 'Union[ast.Module, str]'
-    ) -> 'Dict[str, Set[DataSymbol]]':
+        self,
+        cell: Union[ast.Module, str]
+    ) -> Dict[str, Set[DataSymbol]]:
         if isinstance(cell, str):
             cell = self._get_cell_ast(cell)
         live_symbol_refs, dead_symbol_refs = compute_live_dead_symbol_refs(cell)
@@ -333,7 +360,7 @@ class NotebookSafety(object):
         self._last_refused_code = None
         return False
 
-    def _resync_symbols(self, symbols: 'Set[DataSymbol]'):
+    def _resync_symbols(self, symbols: Set[DataSymbol]):
         for dsym in symbols:
             if not dsym.containing_scope.is_global:
                 continue
@@ -351,7 +378,6 @@ class NotebookSafety(object):
                 containing_obj = containing_scope._obj_ref()
                 # TODO: handle dict case too
                 if isinstance(containing_obj, list) and containing_obj[-1] is obj:
-                    # new_alias_sym = containing_scope.upsert_data_symbol_for_name(
                     containing_scope.upsert_data_symbol_for_name(
                         len(containing_obj) - 1,
                         obj,
@@ -360,7 +386,6 @@ class NotebookSafety(object):
                         is_subscript=True,
                         propagate=False
                     )
-                    # self.aliases[id(obj)].add(new_alias_sym)
             self.aliases[dsym.cached_obj_id].discard(dsym)
             self.aliases[dsym.obj_id].discard(dsym)
             self.aliases[id(obj)].add(dsym)
@@ -368,10 +393,11 @@ class NotebookSafety(object):
             if namespace is not None:
                 namespace.update_obj_ref(obj)
                 del self.namespaces[dsym.obj_id]
-                self.namespaces[id(obj)] = namespace
+                self.namespaces[namespace.obj_id] = namespace
             dsym.update_obj_ref(obj)
 
     def safe_execute(self, cell: str, run_cell_func):
+        ret = None
         with save_number_of_currently_executing_cell():
             self._last_execution_counter = self.cell_counter()
 
@@ -379,13 +405,14 @@ class NotebookSafety(object):
                 self._counters_by_cell_id[self._active_cell_id] = self._last_execution_counter
                 self._active_cell_id = None
             # Stage 1: Precheck.
-            if self._precheck_for_stale(cell) and self.config.get('skip_unsafe_cells', True):
+            if self._precheck_for_stale(cell) and self.settings.get('skip_unsafe_cells', True):
                 # FIXME: hack to increase cell number
                 #  ideally we shouldn't show a cell number at all if we fail precheck since nothing executed
                 return run_cell_func('None')
 
             # Stage 2: Trace / run the cell, updating dependencies as they are encountered.
             try:
+                self.cell_content_by_counter[self._last_execution_counter] = cell
                 with self._tracing_context():
                     ret = run_cell_func(cell)
                 # Stage 2.1: resync any defined symbols that could have gotten out-of-sync
@@ -393,16 +420,20 @@ class NotebookSafety(object):
                 defined = self._check_cell_and_resolve_symbols(cell)['dead']
                 self._resync_symbols(defined)
             finally:
-                if not self.config.store_history:
+                if not self.settings.store_history:
                     self._cell_counter += 1
                 return ret
 
     def _make_cell_magic(self, cell_magic_name):
+        # this is to avoid capturing `self` and creating an extra reference to the singleton
+        store_history = self.settings.store_history
+        self_ = lambda: singletons.nbs()
+
         def _run_cell_func(cell):
-            run_cell(cell, store_history=self.config.store_history)
+            run_cell(cell, store_history=store_history)
 
         def _dependency_safety(_, cell: str):
-            self.safe_execute(cell, _run_cell_func)
+            self_().safe_execute(cell, _run_cell_func)
 
         # FIXME (smacke): probably not a great idea to rely on this
         _dependency_safety.__name__ = cell_magic_name
@@ -415,7 +446,7 @@ class NotebookSafety(object):
         self._recorded_cell_name_to_cell_num = False
 
         try:
-            with self.tracing_manager.tracing_context():
+            with TraceManager.instance().tracing_context():
                 with ast_transformer_context([SafetyAstRewriter(self)]):
                     yield
         finally:
@@ -426,31 +457,33 @@ class NotebookSafety(object):
     def _reset_trace_state_hook(self):
         # this assert doesn't hold anymore now that tracing could be disabled inside of something
         # assert len(self.attr_trace_manager.stack) == 0
-        self.tracing_manager = TracingManager(self)
+        TraceManager.clear_instance()
         self._gc()
 
     def _make_line_magic(self):
         line_magic_names = [f[0] for f in inspect.getmembers(line_magics) if inspect.isfunction(f[1])]
 
         def _safety(line_: str):
+            # this is to avoid capturing `self` and creating an extra reference to the singleton
+            self_ = singletons.nbs()
             line = line_.split()
             if not line or line[0] not in line_magic_names:
                 print(line_magics.USAGE)
                 return
             elif line[0] in ("show_deps", "show_dependency", "show_dependencies"):
-                return line_magics.show_deps(self, line)
+                return line_magics.show_deps(self_, line)
             elif line[0] == "show_stale":
-                return line_magics.show_stale(self, line)
+                return line_magics.show_stale(self_, line)
             elif line[0] == "trace_messages":
-                return line_magics.trace_messages(self, line)
+                return line_magics.trace_messages(self_, line)
             elif line[0] == "remove_dependency":
-                return line_magics.remove_dep(self, line)
+                return line_magics.remove_dep(self_, line)
             elif line[0] in ("add_dependency", "add_dep"):
-                return line_magics.add_dep(self, line)
+                return line_magics.add_dep(self_, line)
             elif line[0] == "turn_off_warnings_for":
-                return line_magics.turn_off_warnings_for(self, line)
+                return line_magics.turn_off_warnings_for(self_, line)
             elif line[0] == "turn_on_warnings_for":
-                return line_magics.turn_on_warnings_for(self, line)
+                return line_magics.turn_on_warnings_for(self_, line)
 
         # FIXME (smacke): probably not a great idea to rely on this
         _safety.__name__ = _SAFETY_LINE_MAGIC
@@ -458,7 +491,7 @@ class NotebookSafety(object):
 
     @property
     def dependency_tracking_enabled(self):
-        return self.config.get('track_dependencies', True)
+        return self.settings.get('track_dependencies', True)
 
     @property
     def cell_magic_name(self):
@@ -482,6 +515,7 @@ class NotebookSafety(object):
         for obj_id in self.garbage_namespace_obj_ids:
             garbage_ns = self.namespaces.pop(obj_id, None)
             if garbage_ns is not None:
+                logger.info("collect ns %s", garbage_ns)
                 garbage_ns.clear_namespace(obj_id)
         self.garbage_namespace_obj_ids.clear()
         # while True:
@@ -497,9 +531,10 @@ class NotebookSafety(object):
     def _gc(self):
         for dsym in list(self.all_data_symbols()):
             if dsym.is_garbage:
+                logger.info("collect sym %s", dsym)
                 dsym.collect_self_garbage()
 
-    def retrieve_namespace_attr_or_sub(self, obj: 'Any', attr_or_sub: 'Union[str, int]', is_subscript: bool):
+    def retrieve_namespace_attr_or_sub(self, obj: Any, attr_or_sub: Union[str, int], is_subscript: bool):
         try:
             if is_subscript:
                 # TODO: more complete list of things that are checkable
