@@ -1,5 +1,5 @@
-# -*- coding: utf-8 -*-
+# -*- coding: future_annotations -*-
 from .safety_ast_rewriter import SafetyAstRewriter
-from .trace_manager import TracingManager
+from .trace_manager import TraceManager
 from .trace_events import TraceEvent
 from .trace_stmt import TraceStatement
