@@ -1,20 +1,35 @@
-# -*- coding: utf-8 -*-
+# -*- coding: future_annotations -*-
+from contextlib import contextmanager
 import logging
 
-import pytest
-
+from nbsafety.safety import NotebookSafetySettings
+from nbsafety.singletons import nbs
 from .utils import make_safety_fixture, skipif_known_failing
 
 logging.basicConfig(level=logging.ERROR)
 
 # Reset dependency graph before each test
-_safety_fixture, _safety_state, run_cell_ = make_safety_fixture()
+# _safety_fixture, run_cell_ = make_safety_fixture(trace_messages_enabled=True)
+_safety_fixture, run_cell_ = make_safety_fixture()
+
+
+@contextmanager
+def override_settings(**kwargs):
+    old_settings = nbs().settings
+    new_settings = old_settings._asdict()
+    new_settings.update(kwargs)
+    new_settings = NotebookSafetySettings(**new_settings)
+    try:
+        nbs().settings = new_settings
+        yield
+    finally:
+        nbs().settings = old_settings
 
 
 def run_cell(cell, cell_id=None, **kwargs):
     """Mocks the `change active cell` portion of the comm protocol"""
     if cell_id is not None:
-        _safety_state[0].handle({
+        nbs().handle({
             'type': 'change_active_cell',
             'active_cell_id': cell_id
         })
@@ -31,7 +46,7 @@ def test_simple():
     run_cell(cells[0])
     run_cell(cells[1])
     run_cell(cells[2])
-    response = _safety_state[0].check_and_link_multiple_cells(cells)
+    response = nbs().check_and_link_multiple_cells(cells)
     assert response['stale_cells'] == [3]
     assert response['fresh_cells'] == []
     assert response['stale_links'] == {3: [1]}
@@ -47,7 +62,7 @@ def test_refresh_after_exception_fixed():
     run_cell(cells[0], 0)
     run_cell(cells[2], 2, ignore_exceptions=True)
     run_cell(cells[1], 1)
-    response = _safety_state[0].check_and_link_multiple_cells(cells)
+    response = nbs().check_and_link_multiple_cells(cells)
     assert response['fresh_cells'] == [2]
 
 
@@ -62,7 +77,7 @@ def test_refresh_after_val_changed():
     run_cell(cells[1], 1)
     run_cell(cells[2], 2)
     run_cell(cells[3], 3)
-    response = _safety_state[0].check_and_link_multiple_cells(cells)
+    response = nbs().check_and_link_multiple_cells(cells)
     assert response['fresh_cells'] == [2]
 
 
@@ -76,16 +91,17 @@ def test_inner_mutation_considered_fresh():
     }
     for idx, cell in cells.items():
         run_cell(cell, idx)
-    response = _safety_state[0].check_and_link_multiple_cells(cells)
+    response = nbs().check_and_link_multiple_cells(cells)
     assert response['stale_cells'] == []
     assert response['fresh_cells'] == [2, 3]
 
 
-@pytest.mark.parametrize("force_subscript_symbol_creation", [True, False])
-def test_update_list_elem(force_subscript_symbol_creation):
+# @pytest.mark.parametrize("force_subscript_symbol_creation", [True, False])
+def test_update_list_elem():
+    force_subscript_symbol_creation = True
     cells = {
         0: """
-class Foo(object):
+class Foo:
     def __init__(self):
         self.counter = 0
         self.dummy = 0
@@ -109,33 +125,33 @@ for foo in lst:
     for idx, cell in cells.items():
         run_cell(cell, idx)
 
-    response = _safety_state[0].check_and_link_multiple_cells(cells)
+    response = nbs().check_and_link_multiple_cells(cells)
     assert response['stale_cells'] == []
     assert response['fresh_cells'] == []
 
     cells[4] = 'x.inc()'
     run_cell(cells[4], 4)
 
-    response = _safety_state[0].check_and_link_multiple_cells(cells)
-    assert response['stale_cells'] == []
+    response = nbs().check_and_link_multiple_cells(cells)
+    assert response['stale_cells'] == [], 'got %s' % response['stale_cells']
     assert response['fresh_cells'] == [2, 3]
 
     cells[5] = 'foo.inc()'
     run_cell(cells[5], 5)
-    response = _safety_state[0].check_and_link_multiple_cells(cells)
+    response = nbs().check_and_link_multiple_cells(cells)
     assert response['stale_cells'] == []
     assert response['fresh_cells'] == [2, 3, 4]
 
     if force_subscript_symbol_creation:
         cells[6] = 'lst[-1]'
         run_cell(cells[6], 6)
-        response = _safety_state[0].check_and_link_multiple_cells(cells)
+        response = nbs().check_and_link_multiple_cells(cells)
         assert response['stale_cells'] == []
         assert response['fresh_cells'] == [2, 3, 4]
 
     run_cell(cells[4], 4)
     run_cell('%safety show_deps foo', 1234)
-    response = _safety_state[0].check_and_link_multiple_cells(cells)
+    response = nbs().check_and_link_multiple_cells(cells)
     assert response['stale_cells'] == []
     assert response['fresh_cells'] == [2, 3, 5] + ([6] if force_subscript_symbol_creation else [])
 
@@ -149,7 +165,7 @@ def test_no_freshness_for_alias_assignment_post_mutation():
     }
     for idx, cell in cells.items():
         run_cell(cell, idx)
-    response = _safety_state[0].check_and_link_multiple_cells(cells)
+    response = nbs().check_and_link_multiple_cells(cells)
     assert response['stale_cells'] == []
     assert response['fresh_cells'] == []
 
@@ -161,7 +177,7 @@ def test_fresh_after_import():
     }
     for idx, cell in cells.items():
         run_cell(cell, idx, ignore_exceptions=True)
-    response = _safety_state[0].check_and_link_multiple_cells(cells)
+    response = nbs().check_and_link_multiple_cells(cells)
     assert response['stale_cells'] == []
     assert response['fresh_cells'] == [0]
 
@@ -176,16 +192,12 @@ def test_external_object_update_propagates_to_stale_namespace_symbols():
         5: 'x = 43',
         6: 'foo = foo.set_x(10)',
     }
-    old_skip_stale = _safety_state[0].config.get('skip_unsafe_cells', True)
-    try:
-        _safety_state[0].config.skip_unsafe_cells = False
+    with override_settings(skip_unsafe_cells=False):
         for idx, cell in cells.items():
             run_cell(cell, idx)
-        response = _safety_state[0].check_and_link_multiple_cells(cells)
+        response = nbs().check_and_link_multiple_cells(cells)
         assert response['stale_cells'] == []
         assert response['fresh_cells'] == [2, 4]
-    finally:
-        _safety_state[0].config.skip_unsafe_cells = old_skip_stale
 
 
 def test_symbol_on_both_sides_of_assignment():
@@ -197,7 +209,34 @@ def test_symbol_on_both_sides_of_assignment():
     for idx, cell in cells.items():
         run_cell(cell, idx)
     cells[3] = 'y += 7'
-    response = _safety_state[0].check_and_link_multiple_cells(cells)
+    response = nbs().check_and_link_multiple_cells(cells)
     assert response['stale_cells'] == [3]
     assert response['fresh_cells'] == [1]
     assert list(response['refresher_links'].keys()) == [1]
+
+
+def test_updated_namespace_after_subscript_dep_removed():
+    cells = {
+        0: 'x = 5',
+        1: 'd = {x: 5}',
+        2: 'logging.info(d[5])',
+        3: 'x = 9',
+    }
+    for idx, cell in cells.items():
+        run_cell(cell, idx)
+    response = nbs().check_and_link_multiple_cells(cells)
+    assert response['stale_cells'] == [2]
+    assert response['fresh_cells'] == [1]
+    cells[1] = 'd = {5: 5}'
+    run_cell(cells[1], 1)
+    response = nbs().check_and_link_multiple_cells(cells)
+    assert response['stale_cells'] == []
+    assert response['fresh_cells'] == [2]
+    run_cell(cells[2], 2)
+    response = nbs().check_and_link_multiple_cells(cells)
+    assert response['stale_cells'] == []
+    assert response['fresh_cells'] == []
+    run_cell(cells[0], 0)
+    response = nbs().check_and_link_multiple_cells(cells)
+    assert response['stale_cells'] == []
+    assert response['fresh_cells'] == [], 'got %s' % response['fresh_cells']
