@@ -1,16 +1,18 @@
-# -*- coding: utf-8 -*-
+# -*- coding: future_annotations -*-
 import logging
 
-from test.utils import make_safety_fixture
+from nbsafety.singletons import nbs
+from .utils import make_safety_fixture
 
 logging.basicConfig(level=logging.ERROR)
 
 # Reset dependency graph before each test
-_safety_fixture, _safety_state, run_cell_ = make_safety_fixture()
+# _safety_fixture, run_cell_ = make_safety_fixture(trace_messages_enabled=True)
+_safety_fixture, run_cell_ = make_safety_fixture()
 
 
 def updated_symbol_names():
-    return sorted(map(lambda sym: sym.readable_name, _safety_state[0].updated_symbols))
+    return sorted([sym.readable_name for sym in nbs().updated_symbols if not sym.is_anonymous])
 
 
 def run_cell(cell):
@@ -31,10 +33,12 @@ def test_simplest():
 
 def test_dict_hierarchy():
     run_cell('d = {}')
-    assert updated_symbol_names() == ['d']
+    updated_sym_names = updated_symbol_names()
+    assert updated_sym_names == ['d'], 'got %s' % updated_sym_names
     run_cell('d["foo"] = {}')
     assert updated_symbol_names() == sorted(['d[foo]', 'd'])
     run_cell('d["foo"]["bar"] = []')
-    assert updated_symbol_names() == sorted(['d[foo][bar]', 'd[foo]', 'd'])
+    updated_sym_names = updated_symbol_names()
+    assert updated_sym_names == sorted(['d[foo][bar]', 'd[foo]', 'd']), 'got %s' % updated_sym_names
     run_cell('d["foo"]["bar"] = 0')
     assert updated_symbol_names() == sorted(['d[foo][bar]', 'd[foo]', 'd'])
